from desgld.desgld_alg import DeSGLD
from desgld.evaluation import ClassificationAccuracy, Wasserstein2Distance
from desgld.network import NetworkArchitecture

__all__ = [
    "NetworkArchitecture",
    "ClassificationAccuracy",
    "Wasserstein2Distance",
    "DeSGLD",
]



import math

import numpy as np
from scipy.linalg import sqrtm


class ClassificationAccuracy:
    """Calculate the classification accuracy in Bayesian Logistic Regression

    Args:
        x_all (list): the input data
        y_all (list): the output data
        history_all (list): contains the approximation from all the nodes
        T (int): the number of iterations
    """

    def __init__(self, x_all, y_all, history_all, T):
        self.x_all = x_all
        self.y_all = y_all
        self.history_all = history_all
        self.T = T

    def compute_accuracy(self):
        """Function that computes the classification accuracy

        Args:
            params: the same parameters from the class

        Returns:
        result_acc (float): the classification accuracy
        result_std (float): the standard deviation of the classification
        accuracy
        """
        mis_class = np.empty((self.T + 1, len(self.history_all[0, 0, 0])))

        for t in range(self.T + 1):
            for n in range(len(self.history_all[t, 0, 0])):
                temp0 = 0
                for i in range(len(self.x_all)):
                    z = 1 / (
                        1
                        + np.exp(
                            -np.dot(
                                np.transpose(self.history_all[t, 1])[n],
                                self.x_all[i],
                            )
                        )
                    )
                    if z >= 0.5:
                        z = 1
                    else:
                        z = 0
                    if self.y_all[i] != z:
                        temp0 += 1
                mis_class[t, n] = 1 - temp0 / len(self.x_all)

        result_acc = np.mean(mis_class, axis=1)
        result_std = np.std(mis_class, axis=1)
        return result_acc, result_std


class Wasserstein2Distance:
    """Class for: Wasserstein 2 distance in Bayesian Linear Regression

    Args:
        size_w (int): the size of the network
        T (int): the number of iterations
        avg_post (list): the mean of the posterior distribution
        cov_post (list): the covariance of the posterior distribution
        history_all (list): contains the approximation from all the nodes
        beta_mean_all (list): contains the mean of the approximation
        from all the nodes
    """

    def __init__(
        self, size_w, T, avg_post, cov_post, history_all, beta_mean_all
    ):
        self.size_w = size_w
        self.T = T
        self.avg_post = avg_post
        self.cov_post = cov_post
        self.history_all = history_all
        self.beta_mean_all = beta_mean_all

    def W2_dist(self):
        """Class for: Wasserstein 2 distance in Bayesian Linear Regression

        Args:
            size_w (int): the size of the network
            T (int): the number of iterations
            avg_post (list): the mean of the posterior distribution
            cov_post (list): the covariance of the posterior distribution
            history_all (list): contains the approximation from all the nodes
            beta_mean_all (list): contains the mean of the approximation
            from all the nodes
        Returns:
        w2dis (list): contains the W2 distance of each agent and
        the mean of the approximation from all the ageents
        """
        w2dis = []
        for i in range(self.size_w):
            temp = []
            w2dis.append(temp)
        temp = []
        w2dis.append(temp)
        """
        W2 distance of each agent
        """
        for i in range(self.size_w):
            for t in range(self.T + 1):
                d = 0
                avg_temp = []
                avg_temp.append(np.mean(self.history_all[t][i][0]))
                avg_temp.append(np.mean(self.history_all[t][i][1]))
                avg_temp = np.array(avg_temp)
                cov_temp = np.cov(self.history_all[t][i])
                d = np.linalg.norm(self.avg_post - avg_temp) * np.linalg.norm(
                    self.avg_post - avg_temp
                )
                d = d + np.trace(
                    self.cov_post
                    + cov_temp
                    - 2
                    * sqrtm(
                        np.dot(
                            np.dot(sqrtm(cov_temp), self.cov_post),
                            sqrtm(cov_temp),
                        )
                    )
                )
                w2dis[i].append(np.array(math.sqrt(abs(d))))
        """
        W2 distance of the mean of agents
        """
        for t in range(self.T + 1):
            d = 0
            avg_temp = []
            avg_temp.append(np.mean(self.beta_mean_all[t][0]))
            avg_temp.append(np.mean(self.beta_mean_all[t][1]))
            avg_temp = np.array(avg_temp)
            cov_temp = np.cov(self.beta_mean_all[t])
            d = np.linalg.norm(self.avg_post - avg_temp) * np.linalg.norm(
                self.avg_post - avg_temp
            )
            d = d + np.trace(
                self.cov_post
                + cov_temp
                - 2
                * sqrtm(
                    np.dot(
                        np.dot(sqrtm(cov_temp), self.cov_post), sqrtm(cov_temp)
                    )
                )
            )
            w2dis[self.size_w].append(np.array(math.sqrt(abs(d))))

        for i in range(len(w2dis)):
            w2dis[i] = np.array(w2dis[i])

        return w2dis



import math

import numpy as np


class DeSGLD:
    """Decentralized Stochastic Gradient Langevin Dynamics

    Args:
        size_w (int): the size of the network
        N (int): the size of the array in each it
        sigma (float): the variation of the noise
        eta (float): the learning rate
        T (int): the iteration numbers
        dim (int): the dimension of the input data
        b (int): the batch size
        lam (float): the regularization parameter
        x (float): the input data
        y (float): the output data
        w (float): the weight matrix from the network structure
        hv (float): the tuning parameter for the extra algorithm
        reg_type (str): the type of regressor used

    Returns:
        vanila_desgld()
        extra_desgld()
    """

    def __init__(
        self, size_w, N, sigma, eta, T, dim, b, lam, x, y, w, hv, reg_type
    ):
        self.size_w = size_w
        self.N = N
        self.sigma = sigma
        self.eta = eta
        self.dim = dim
        self.b = b
        self.lam = lam
        self.x = x
        self.y = y
        self.w = w
        self.hv = hv
        self.T = T
        self.reg_type = reg_type

    def gradient_logreg(self, beta, x, y, dim, lam, b):
        """Gradient function for the logistic regression

        Args:
            beta: approximation at each node
            x: the input data
            y: the output data
            dim: the dimension of the input data
            lam: the regularization parameter
            b: the batch size
        Returns:
            gradient for the the logistic regression
        """
        f = np.zeros(dim)
        randomList = np.random.randint(0, len(y) - 1, size=int(b))
        for item in randomList:
            h = 1 / (1 + np.exp(-np.dot(beta, x[item])))
            f -= np.dot((y[item] - h), x[item])
        f += (2 / lam) * beta
        return f

    def gradient_linreg(self, beta, x, y, dim, lam, b):
        """Gradient function for the linear regression

        Args:
            beta: approximation at each node
            x: the input data
            y: the output data
            dim: the dimension of the input data
            lam: the regularization parameter
            b: the batch size
        Returns:
            gradient for the the logistic regression
        """
        f = np.zeros(dim)
        randomList = np.random.randint(0, len(y) - 1, size=int(b))
        for i in randomList:
            f = f - np.dot((y[i] - np.dot(beta, x[i])), x[i])
        f += (2 / lam) * beta
        return f

    def vanila_desgld(self):
        """Vanila DeSGLD algorithm

        Returns:
            history_all: contains the approximation from all the nodes
            beta_mean_all: contains the mean of the approximation from
            all the nodes
        """
        # Initialization
        if self.reg_type == "logistic":
            beta = np.random.normal(
                0, self.sigma, size=(self.N, self.size_w, self.dim)
            )
        else:
            beta = np.random.multivariate_normal(
                mean=np.zeros(self.dim),
                cov=np.eye(self.dim),
                size=(self.N, self.size_w),
            )
        history_all = []
        beta_mean_all = []
        for t in range(1):
            history = np.empty((self.size_w, self.dim, self.N))
            beta_mean = np.empty((self.dim, self.N))
            for i in range(self.N):
                history[:, :, i] = beta[i, :, :]
            for j in range(self.dim):
                beta_mean[j, :] = np.mean(history[:, j, :], axis=0)
            history_all.append(history)
            beta_mean_all.append(beta_mean)

        # Update
        step = self.eta
        for t in range(self.T):
            for n in range(self.N):
                for i in range(self.size_w):
                    if self.reg_type == "logistic":
                        g = self.gradient_logreg(
                            beta[n, i],
                            self.x[i],
                            self.y[i],
                            self.dim,
                            self.lam,
                            self.b,
                        )
                        temp = np.zeros(self.dim)
                        for j in range(len(beta[n])):
                            temp = temp + self.w[i, j] * beta[n, j]
                        noise = np.random.normal(0, self.sigma, self.dim)
                        beta[n, i] = (
                            temp - step * g + math.sqrt(2 * step) * noise
                        )
                    else:
                        g = self.gradient_linreg(
                            beta[n, i],
                            self.x[i],
                            self.y[i],
                            self.dim,
                            self.lam,
                            self.b,
                        )
                        temp = np.zeros(self.dim)
                        for j in range(len(beta[n])):
                            temp = temp + self.w[i, j] * beta[n, j]
                        noise = np.random.multivariate_normal(
                            mean=np.zeros(self.dim), cov=np.eye(self.dim)
                        )
                        beta[n, i] = (
                            temp - step * g + math.sqrt(2 * step) * noise
                        )

            history = np.empty((self.size_w, self.dim, self.N))
            beta_mean = np.empty((self.dim, self.N))
            for i in range(self.N):
                history[:, :, i] = beta[i, :, :]
            for j in range(self.dim):
                beta_mean[j, :] = np.mean(history[:, j, :], axis=0)
            history_all.append(history)
            beta_mean_all.append(beta_mean)
        return np.array(history_all), np.array(beta_mean_all)

    def extra_desgld(self):
        """EXTRA DeSGLD algorithm

        Returns:
            history_all: contains the approximation from all the nodes
            beta_mean_all: contains the mean of the approximation from
            all the nodes
        """
        # I_n is the identity matrix of the same size as the weight matrix
        I_n = np.eye(self.size_w)
        h_values = self.hv

        history_all = []
        beta_mean_all = []
        for h in h_values:
            # w1 is the weight matrix with the hyperparameter h
            w1 = h * I_n + (1 - h) * self.w

            # Initialization
            if self.reg_type == "logistic":
                beta = np.random.normal(
                    0, self.sigma, size=(self.N, self.size_w, self.dim)
                )
            else:
                beta = np.random.multivariate_normal(
                    mean=np.zeros(self.dim),
                    cov=np.eye(self.dim),
                    size=(self.N, self.size_w),
                )
            history_all_h = []
            beta_mean_all_h = []
            for t in range(1):
                history = np.empty((self.size_w, self.dim, self.N))
                beta_mean = np.empty((self.dim, self.N))
                for i in range(self.N):
                    history[:, :, i] = beta[i, :, :]
                for j in range(self.dim):
                    beta_mean[j, :] = np.mean(history[:, j, :], axis=0)
                history_all_h.append(history)
                beta_mean_all_h.append(beta_mean)

            # Update
            step = self.eta
            for t in range(self.T):
                for n in range(self.N):
                    for i in range(self.size_w):
                        if self.reg_type == "logistic":
                            # Vanila Part
                            g = self.gradient_logreg(
                                beta[n, i],
                                self.x[i],
                                self.y[i],
                                self.dim,
                                self.lam,
                                self.b,
                            )
                            temp = np.zeros(self.dim)
                            for j in range(len(beta[n])):
                                temp = temp + w1[i, j] * beta[n, j]
                            noise = np.random.normal(0, self.sigma, self.dim)
                            beta[n, i] = (
                                temp - step * g + math.sqrt(2 * step) * noise
                            )
                            # Extra Part
                            g = self.gradient_logreg(
                                beta[n, i],
                                self.x[i],
                                self.y[i],
                                self.dim,
                                self.lam,
                                self.b,
                            )
                            temp = np.zeros(self.dim)
                            for j in range(len(beta[n])):
                                temp = temp + self.w[i, j] * beta[n, j]
                            noise = np.random.normal(0, self.sigma, self.dim)
                            beta[n, i] = (
                                temp - step * g + math.sqrt(2 * step) * noise
                            )
                        else:
                            # Vanila Part
                            g = self.gradient_linreg(
                                beta[n, i],
                                self.x[i],
                                self.y[i],
                                self.dim,
                                self.lam,
                                self.b,
                            )
                            temp = np.zeros(self.dim)
                            for j in range(len(beta[n])):
                                temp = temp + w1[i, j] * beta[n, j]
                            noise = np.random.multivariate_normal(
                                mean=np.zeros(self.dim), cov=np.eye(self.dim)
                            )
                            beta[n, i] = (
                                temp - step * g + math.sqrt(2 * step) * noise
                            )
                            # Extra Part
                            g = self.gradient_linreg(
                                beta[n, i],
                                self.x[i],
                                self.y[i],
                                self.dim,
                                self.lam,
                                self.b,
                            )
                            temp = np.zeros(self.dim)
                            for j in range(len(beta[n])):
                                temp = temp + self.w[i, j] * beta[n, j]
                            noise = np.random.multivariate_normal(
                                mean=np.zeros(self.dim), cov=np.eye(self.dim)
                            )
                            beta[n, i] = (
                                temp - step * g + math.sqrt(2 * step) * noise
                            )

                history = np.empty((self.size_w, self.dim, self.N))
                beta_mean = np.empty((self.dim, self.N))
                for i in range(self.N):
                    history[:, :, i] = beta[i, :, :]
                for j in range(self.dim):
                    beta_mean[j, :] = np.mean(history[:, j, :], axis=0)
                history_all_h.append(history)
                beta_mean_all_h.append(beta_mean)
            history_all.append(history_all_h)
            beta_mean_all.append(beta_mean_all_h)
        return np.array(history_all), np.array(beta_mean_all)



import random

import numpy as np


class NetworkArchitecture:
    """Decentralized network architecture

    Description:
        This class is used to generate different network
        architectures. Here we add four different network architectures:
        fully connected, fully disconnected, circular network and star network.
    Args:
        size_w (int): the size of the network
        random_seed (int or None): random seed for reproducibility
    Returns:
        Different network architectures
    """

    def __init__(self, size_w, random_seed=None):
        self.size_w = size_w
        self.random_seed = random_seed
        self.set_seed()

    def set_seed(self):
        if self.random_seed is not None:
            np.random.seed(self.random_seed)

    def fully_connected(self):
        """Fully Connected Network

        Description:
            A fully connected network is a kind of network in which all nodes
            are connected to all other nodes.

        Args:
            deg_mat(np.ndarray): degree matrix of the graph Laplacian
            adj_mat(np.ndarray): adjacency matrix of the graph Laplacian
            laplacian_matrix(np.ndarray): graph Laplacian matrix
            eigenvalues (np.ndarray): eigenvalues of the graph Laplacian
            max_eigenvalue (float): maximum eigenvalue of the graph Laplacian
            delta (float): delta of the graph Laplacian


        Examples:
            >>> w=NetworkArchitecture(size_w=5,random_seed=42)
            >>> w=w.fully_connected()
            >>> print(w)
                [[0.67811004 0.08047249 0.08047249 0.08047249 0.08047249]
                [0.08047249 0.67811004 0.08047249 0.08047249 0.08047249]
                [0.08047249 0.08047249 0.67811004 0.08047249 0.08047249]
                [0.08047249 0.08047249 0.08047249 0.67811004 0.08047249]
                [0.08047249 0.08047249 0.08047249 0.08047249 0.67811004]]
            >>> print("Row sums:", np.sum(w, axis=0))
                Row sums: [1. 1. 1. 1. 1.]
            >>> print("Column sums:", np.sum(w, axis=1))
                Column sums: [1. 1. 1. 1. 1.]
        Returns:
            w (float): a fully connected network matrix
        """
        adj_mat = np.ones((self.size_w, self.size_w)) - np.eye(self.size_w)
        deg_mat = (self.size_w - 1) * np.eye(self.size_w)
        laplacian_matrix = deg_mat - adj_mat
        eigenvalues = np.linalg.eigvals(laplacian_matrix)
        max_eigenvalue = np.max(np.real(eigenvalues))
        delta = random.uniform(0, (2 / max_eigenvalue))
        w = np.eye(self.size_w) - delta * laplacian_matrix
        w = np.abs(w)
        row_sums = np.sum(w, axis=1)
        w = w / row_sums[:, np.newaxis]

        return w

    def circular_network(self):
        """Circular Network

        Description:
            A circular network is a kind of network in which a particular
            node is connected to its left and right nodes only.

        Args:
            deg_mat(np.ndarray): degree matrix of the graph Laplacian
            adj_mat(np.ndarray): adjacency matrix of the graph Laplacian
            laplacian_matrix(np.ndarray): graph Laplacian matrix
            eigenvalues (np.ndarray): eigenvalues of the graph Laplacian
            max_eigenvalue (float): maximum eigenvalue of the graph Laplacian
            delta (float): delta of the graph Laplacian


        Returns:
            w (float): a circular network matrix

        Examples:
            >>> w=NetworkArchitecture(size_w=5,random_seed=42)
            >>> w=net.circular_network()
            >>> print(w)
                [[0.72162653 0.13918674 0.         0.         0.13918674]
                [0.13918674 0.72162653 0.13918674 0.         0.        ]
                [0.         0.13918674 0.72162653 0.13918674 0.        ]
                [0.         0.         0.13918674 0.72162653 0.13918674]
                [0.13918674 0.         0.         0.13918674 0.72162653]]
            >>> print("Row sums:", np.sum(w, axis=0))
                Row sums: [1. 1. 1. 1. 1.]
            >>> print("Column sums:", np.sum(w, axis=1))
                Column sums: [1. 1. 1. 1. 1.]
        """
        adj_mat = np.zeros((self.size_w, self.size_w))
        for i in range(self.size_w):
            for j in range(self.size_w):
                if (i + 1) == j:
                    adj_mat[i, j] = 1
                else:
                    adj_mat[i, j] = adj_mat[i, j]
        adj_mat[0, (self.size_w - 1)] = 1
        adj_mat = adj_mat + np.transpose(adj_mat)
        deg_mat = 2 * np.eye(self.size_w)
        laplacian_matrix = deg_mat - adj_mat
        eigenvalues = np.linalg.eigvals(laplacian_matrix)
        max_eigenvalue = np.max(np.real(eigenvalues))
        delta = random.uniform(0, (2 / max_eigenvalue))
        w = np.eye(self.size_w) - delta * laplacian_matrix
        w = np.abs(w)
        row_sums = np.sum(w, axis=1)
        w = w / row_sums[:, np.newaxis]

        return w

    def fully_disconnected(self):
        """Completely Disconnected Network

        Description:
            A completely disconnected network is a kind of network in which
            all the nodes are disconnected from each other.

        Returns:
            w (float): a disconnected network matrix

        Examples:
            >>> net = NetworkArchitecture(size_w=6)
            >>> w=net.circular()
            >>> print(w)
                [[1,0,0,0,0,0],
                [0,1,0,0,0,0],
                [0,0,1,0,0,0],
                [0,0,0,1,0,0],
                [0,0,0,0,1,0],
                [0,0,0,0,0,1]]

        """
        x = [[0] * self.size_w for _ in range(self.size_w)]
        for i in range(self.size_w):
            for j in range(self.size_w):
                if i == j:
                    x[i][j] = 1
        return np.array(x)

    def star_network(self):
        """Star-like Connected Network

        Description:
            A star-like network is a kind of network in which
            there is a central node and all other nodes are
            connected to the central node. However, the individual
            nodes are not connected to each other.

        Args:
            deg_mat(np.ndarray): degree matrix of the graph Laplacian
            adj_mat(np.ndarray): adjacency matrix of the graph Laplacian
            laplacian_matrix(np.ndarray): graph Laplacian matrix
            eigenvalues (np.ndarray): eigenvalues of the graph Laplacian
            max_eigenvalue (float): maximum eigenvalue of the graph Laplacian
            delta (float): delta of the graph Laplacian

        Returns:
            w (float): a star network matrix

        Examples:
            >>> w=NetworkArchitecture(size_w=5,random_seed=42)
            >>> w=net.star_network()
            >>> print(w)
                [[0.71180893 0.07204777 0.07204777 0.07204777 0.07204777]
                [0.07204777 0.92795223 0.         0.         0.        ]
                [0.07204777 0.         0.92795223 0.         0.        ]
                [0.07204777 0.         0.         0.92795223 0.        ]
                [0.07204777 0.         0.         0.         0.92795223]]
            >>> print("Row sums:", np.sum(w, axis=0))
                Row sums: [1. 1. 1. 1. 1.]
            >>> print("Column sums:", np.sum(w, axis=1))
                Column sums: [1. 1. 1. 1. 1.]
        """
        adj_mat = np.zeros((self.size_w, self.size_w))
        for i in range(self.size_w):
            for j in range(self.size_w):
                if i == 0 or j == 0:
                    adj_mat[i, j] = 1
                    adj_mat[i, i] = 0
                else:
                    adj_mat[i, j] = adj_mat[i, j]
        deg_mat = np.eye(self.size_w)
        deg_mat[0, 0] = self.size_w - 1
        laplacian_matrix = deg_mat - adj_mat
        eigenvalues = np.linalg.eigvals(laplacian_matrix)
        max_eigenvalue = np.max(np.real(eigenvalues))
        delta = random.uniform(0, (2 / max_eigenvalue))
        w = np.eye(self.size_w) - delta * laplacian_matrix
        w = np.abs(w)
        row_sums = np.sum(w, axis=1)
        w = w / row_sums[:, np.newaxis]

        return w



from setuptools import setup

requirements = [
    "numpy",
    "scipy",
    "scikit-learn",
    "matplotlib",
    "tqdm",
]


requirements_dev = ["black", "isort", "flake8", "pre-commit"]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="desgld",
    version="0.1.6",
    description="Decentralized stochastic gradient Langevin diffusion",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mrislambd/desgld_package.git",
    author="Rafiq Islam",
    packages=["desgld"],
    package_dir={"": "src"},
    install_requires=requirements,
    extras_require={
        "dev": requirements_dev,
    },
)
